@@ -5,8 +5,9 @@ from shapely.geometry import Polygon
 from .. import bounds, geometry, graph, grouping
 from ..constants import log
 from ..constants import tol_path as tol
+from ..iteration import reduce_cascade
 from ..transformations import transform_points
-from ..typed import Iterable, NDArray, Number, Optional, Union, float64
+from ..typed import ArrayLike, Iterable, NDArray, Number, Optional, Union, float64, int64
 from .simplify import fit_circle_check
 from .traversal import resample_path
 
@@ -50,6 +51,16 @@ def enclosure_tree(polygons):
        contained by another polygon
     """
 
+    # nodes are indexes in polygons
+    contains = nx.DiGraph()
+
+    if len(polygons) == 0:
+        return np.array([], dtype=np.int64), contains
+    elif len(polygons) == 1:
+        # add an early exit for only a single polygon
+        contains.add_node(0)
+        return np.array([0], dtype=np.int64), contains
+
     # get the bounds for every valid polygon
     bounds = {
         k: v
@@ -59,8 +70,6 @@ def enclosure_tree(polygons):
         if len(v) == 4
     }
 
-    # nodes are indexes in polygons
-    contains = nx.DiGraph()
     # make sure we don't have orphaned polygon
     contains.add_nodes_from(bounds.keys())
 
@@ -112,17 +121,17 @@ def enclosure_tree(polygons):
     return roots, contains
 
 
-def edges_to_polygons(edges, vertices):
+def edges_to_polygons(edges: NDArray[int64], vertices: NDArray[float64]):
     """
     Given an edge list of indices and associated vertices
     representing lines, generate a list of polygons.
 
     Parameters
     -----------
-    edges : (n, 2) int
+    edges : (n, 2)
       Indexes of vertices which represent lines
-    vertices : (m, 2) float
-      Vertices in 2D space
+    vertices : (m, 2)
+      Vertices in 2D space.
 
     Returns
     ----------
@@ -154,17 +163,17 @@ def edges_to_polygons(edges, vertices):
     # find which polygons contain which other polygons
     roots, tree = enclosure_tree(polygons)
 
-    # generate list of polygons with proper interiors
-    complete = []
-    for root in roots:
-        interior = list(tree[root].keys())
-        shell = polygons[root].exterior.coords
-        holes = [polygons[i].exterior.coords for i in interior]
-        complete.append(Polygon(shell=shell, holes=holes))
-    return complete
+    # generate polygons with proper interiors
+    return [
+        Polygon(
+            shell=polygons[root].exterior,
+            holes=[polygons[i].exterior for i in tree[root].keys()],
+        )
+        for root in roots
+    ]
 
 
-def polygons_obb(polygons: Iterable[Polygon]):
+def polygons_obb(polygons: Union[Iterable[Polygon], ArrayLike]):
     """
     Find the OBBs for a list of shapely.geometry.Polygons
     """
@@ -285,7 +294,7 @@ def plot(polygon=None, show=True, axes=None, **kwargs):
     **kwargs
       Passed to plt.plot
     """
-    import matplotlib.pyplot as plt
+    import matplotlib.pyplot as plt  # noqa
 
     def plot_single(single):
         axes.plot(*single.exterior.xy, **kwargs)
@@ -394,7 +403,7 @@ def medial_axis(polygon: Polygon, resolution: Optional[Number] = None, clip=None
     # a circle will have a single point medial axis
     if len(polygon.interiors) == 0:
         # what is the approximate scale of the polygon
-        scale = np.reshape(polygon.bounds, (2, 2)).ptp(axis=0).max()
+        scale = np.ptp(np.reshape(polygon.bounds, (2, 2)), axis=0).max()
         # a (center, radius, error) tuple
         fit = fit_circle_check(polygon.exterior.coords, scale=scale)
         # is this polygon in fact a circle
@@ -413,7 +422,7 @@ def medial_axis(polygon: Polygon, resolution: Optional[Number] = None, clip=None
     from shapely import vectorized
 
     if resolution is None:
-        resolution = np.reshape(polygon.bounds, (2, 2)).ptp(axis=0).max() / 100
+        resolution = np.ptp(np.reshape(polygon.bounds, (2, 2)), axis=0).max() / 100
 
     # get evenly spaced points on the polygons boundaries
     samples = resample_boundaries(polygon=polygon, resolution=resolution, clip=clip)
@@ -442,7 +451,7 @@ def medial_axis(polygon: Polygon, resolution: Optional[Number] = None, clip=None
 
     if tol.strict:
         # make sure we didn't screw up indexes
-        assert (vertices[edges_final] - voronoi.vertices[edges]).ptp() < 1e-5
+        assert np.ptp(vertices[edges_final] - voronoi.vertices[edges]) < 1e-5
 
     return edges_final, vertices
 
@@ -519,7 +528,7 @@ def polygon_scale(polygon):
     scale : float
       Length of AABB diagonal
     """
-    extents = np.reshape(polygon.bounds, (2, 2)).ptp(axis=0)
+    extents = np.ptp(np.reshape(polygon.bounds, (2, 2)), axis=0)
     scale = (extents**2).sum() ** 0.5
 
     return scale
@@ -551,20 +560,27 @@ def paths_to_polygons(paths, scale=None):
             # non-zero area
             continue
         try:
-            polygons[i] = repair_invalid(Polygon(path), scale)
+            polygon = Polygon(path)
+            if polygon.is_valid:
+                polygons[i] = polygon
+            else:
+                polygons[i] = repair_invalid(polygon, scale)
         except ValueError:
             # raised if a polygon is unrecoverable
             continue
         except BaseException:
             log.error("unrecoverable polygon", exc_info=True)
     polygons = np.array(polygons)
+
     return polygons
 
 
 def sample(polygon, count, factor=1.5, max_iter=10):
     """
     Use rejection sampling to generate random points inside a
-    polygon.
+    polygon. Note that this function may return fewer or no
+    points, in particular if the polygon as very little area
+    compared to the area of the axis-aligned bounding box.
 
     Parameters
     -----------
@@ -587,9 +603,13 @@ def sample(polygon, count, factor=1.5, max_iter=10):
     # do batch point-in-polygon queries
     from shapely import vectorized
 
+    # TODO : this should probably have some option to
+    # sample from the *oriented* bounding box which would
+    # make certain cases much, much more efficient.
+
     # get size of bounding box
     bounds = np.reshape(polygon.bounds, (2, 2))
-    extents = bounds.ptp(axis=0)
+    extents = np.ptp(bounds, axis=0)
 
     # how many points to check per loop iteration
     per_loop = int(count * factor)
@@ -607,8 +627,7 @@ def sample(polygon, count, factor=1.5, max_iter=10):
     # if we have to do iterations loop here slowly
     for _ in range(max_iter):
         # generate points inside polygons AABB
-        points = np.random.random((per_loop, 2))
-        points = (points * extents) + bounds[0]
+        points = (np.random.random((per_loop, 2)) * extents) + bounds[0]
         # do the point in polygon test and append resulting hits
         mask = vectorized.contains(polygon, *points.T)
         hit.append(points[mask])
@@ -663,7 +682,7 @@ def repair_invalid(polygon, scale=None, rtol=0.5):
         return basic
 
     if scale is None:
-        distance = 0.002 * np.reshape(polygon.bounds, (2, 2)).ptp(axis=0).mean()
+        distance = 0.002 * np.ptp(np.reshape(polygon.bounds, (2, 2)), axis=0).mean()
     else:
         distance = 0.002 * scale
 
@@ -714,8 +733,9 @@ def projected(
     ignore_sign=True,
     rpad=1e-5,
     apad=None,
-    tol_dot=0.01,
-    max_regions=200,
+    tol_dot=1e-10,
+    precise: bool = False,
+    precise_eps: float = 1e-10,
 ):
     """
     Project a mesh onto a plane and then extract the polygon
@@ -734,8 +754,6 @@ def projected(
     ----------
     mesh : trimesh.Trimesh
       Source geometry
-    check : bool
-      If True make sure is flat
     normal : (3,) float
       Normal to extract flat pattern along
     origin : None or (3,) float
@@ -754,10 +772,10 @@ def projected(
       and then de-padding result by to avoid zero-width gaps.
     tol_dot : float
       Tolerance for discarding on-edge triangles.
-    max_regions : int
-      Raise an exception if the mesh has more than this
-      number of disconnected regions to fail quickly before
-      unioning.
+    precise : bool
+      Use the precise projection computation using shapely.
+    precise_eps : float
+      Tolerance for precise triangle checks.
 
     Returns
     ----------
@@ -804,21 +822,42 @@ def projected(
     adjacency_check = side[mesh.face_adjacency].all(axis=1)
     adjacency = mesh.face_adjacency[adjacency_check]
 
-    # a sequence of face indexes that are connected
-    face_groups = graph.connected_components(adjacency, nodes=np.nonzero(side)[0])
-
-    # if something is goofy we may end up with thousands of
-    # regions that do nothing except hang for an hour then segfault
-    if len(face_groups) > max_regions:
-        raise ValueError("too many disconnected groups!")
-
-    # reshape edges into shape length of faces for indexing
-    edges = mesh.edges_sorted.reshape((-1, 6))
     # transform from the mesh frame in 3D to the XY plane
     to_2D = geometry.plane_transform(origin=origin, normal=normal)
     # transform mesh vertices to 2D and clip the zero Z
     vertices_2D = transform_points(mesh.vertices, to_2D)[:, :2]
 
+    if precise:
+        # precise mode just unions triangles as one shapely
+        # polygon per triangle which historically has been very slow
+        # but it is more defensible intellectually
+        faces = mesh.faces[side]
+        # round the 2D vertices with slightly more precision
+        # than our final dilate-erode cleanup
+        digits = int(np.abs(np.log10(precise_eps)) + 2)
+        rounded = np.round(vertices_2D, digits)
+        # get the triangles as closed 4-vertex polygons
+        triangles = rounded[np.column_stack((faces, faces[:, :1]))]
+        # do a check for exactly-degenerate triangles where any two
+        # vertices are exactly identical which means the triangle has
+        # zero area
+        valid = ~(triangles[:, [0, 0, 2]] == triangles[:, [1, 2, 1]]).all(axis=2).any(
+            axis=1
+        )
+        # union the valid triangles and then dilate-erode to clean up
+        # any holes or defects smaller than precise_eps
+        return (
+            ops.unary_union([Polygon(f) for f in triangles[valid]])
+            .buffer(precise_eps)
+            .buffer(-precise_eps)
+        )
+
+    # a sequence of face indexes that are connected
+    face_groups = graph.connected_components(adjacency, nodes=np.nonzero(side)[0])
+
+    # reshape edges into shape length of faces for indexing
+    edges = mesh.edges_sorted.reshape((-1, 6))
+
     polygons = []
     for faces in face_groups:
         # index edges by face then shape back to individual edges
@@ -834,34 +873,22 @@ def projected(
         padding += float(apad)
     if rpad is not None:
         # get the 2D scale as the longest side of the AABB
-        scale = vertices_2D.ptp(axis=0).max()
+        scale = np.ptp(vertices_2D, axis=0).max()
         # apply the scale-relative padding
         padding += float(rpad) * scale
 
-    # some types of errors will lead to a bajillion disconnected
-    # regions and the union will take forever to fail
-    # so exit here early
-    if len(polygons) > max_regions:
-        raise ValueError("too many disconnected groups!")
-
     # if there is only one region we don't need to run a union
     elif len(polygons) == 1:
         return polygons[0]
     elif len(polygons) == 0:
         return None
-    else:
-        # inflate each polygon before unioning to remove zero-size
-        # gaps then deflate the result after unioning by the same amount
-        # note the following provides a 25% speedup but needs
-        # more testing to see if it deflates to a decent looking
-        # result:
-        # polygon = ops.unary_union(
-        #    [p.buffer(padding,
-        #              join_style=2,
-        #              mitre_limit=1.5)
-        #     for p in polygons]).buffer(-padding)
-        polygon = ops.unary_union([p.buffer(padding) for p in polygons]).buffer(-padding)
-    return polygon
+
+    # in my tests this was substantially faster than `shapely.ops.unary_union`
+    reduced = reduce_cascade(lambda a, b: a.union(b), polygons)
+
+    # can be None
+    if reduced is not None:
+        return reduced.buffer(padding).buffer(-padding)
 
 
 def second_moments(polygon: Polygon, return_centered=False):
